package be.iminds.ilabt.jfed.experimenter_gui.slice.espec.model;

import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.util.ESpecLogListener;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentState;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import javafx.application.Platform;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

/**
 * ESpecLogModel
 */
public class ESpecLogModel implements ESpecLogListener {
    private final ObservableList<ESpecLogItem> eSpecLogItems;
    private final ObservableList<ESpecStateItem> eSpecStateItems;

    public ESpecLogModel(Experiment experiment) {
        this.eSpecLogItems = FXCollections.observableArrayList();
        this.eSpecStateItems = FXCollections.observableArrayList();

        experiment.addESpecLogListener(this);
        experiment.experimentStateProperty().addListener(
                (observableValue, oldVal, newVal) -> {
                    boolean success = newVal != ExperimentState.FAILING && newVal != ExperimentState.FAILED;
                    addESpecLogItem("Reached State "+newVal, success);
                    addESpecStatusItem(new ESpecStateItem(newVal+"",
                            success ? (
                                    newVal.isActiveState() ?
                                            ESpecStateItem.ESpecStateStatus.RUNNING :
                                            ESpecStateItem.ESpecStateStatus.SUCCESS) :
                                    ESpecStateItem.ESpecStateStatus.FAILURE)
                    );
                }
        );
    }

    public ObservableList<ESpecLogItem> geteSpecLogItems() {
        return eSpecLogItems;
    }

    public ObservableList<ESpecStateItem> geteSpecStateItems() {
        return eSpecStateItems;
    }



    public boolean isFailingState(@Nonnull String stateName) {
        return stateName.equalsIgnoreCase(ExperimentState.FAILING.name()) ||
                stateName.equalsIgnoreCase(ExperimentState.FAILED.name());
    }


    public void addESpecStatusItem(@Nonnull ESpecStateItem eSpecStateItem) {
        Platform.runLater(
                () -> {
                    if (!eSpecStateItems.isEmpty()) {
                        ESpecStateItem lastStateItem = eSpecStateItems.get(eSpecStateItems.size()-1);
                        if (isFailingState(lastStateItem.getName()))
                            lastStateItem.setStatus(ESpecStateItem.ESpecStateStatus.FAILURE);
                        else
                            lastStateItem.setStatus(ESpecStateItem.ESpecStateStatus.SUCCESS);
                    }
                    eSpecStateItems.add(eSpecStateItem);
                }
        );
    }
    public void addSubESpecStatusItem(@Nonnull String baseStateName,
                                      @Nonnull ESpecStateItem subESpecStateItem) {
        Platform.runLater(() -> {
            Optional<ESpecStateItem> parent = eSpecStateItems
                    .stream()
                    .filter(e -> Objects.equals(e.getName(), baseStateName))
                    .findFirst();
            if (!parent.isPresent()) {
                //TODO: handle case of sub status arriving after failure which prevented parent creation
                //      by adding before the failure in that case?
                parent = Optional.of(new ESpecStateItem(baseStateName));
                eSpecStateItems.add(parent.get());
            }
            parent.get().getChildren().add(subESpecStateItem);
        });
    }
    public void setSubStatus(@Nonnull String baseStateName,
                             @Nonnull String subStateName,
                             @Nonnull ESpecStateItem.ESpecStateStatus newStatus) {
        Optional<ESpecStateItem> parent = eSpecStateItems
                            .stream()
                            .filter(e -> Objects.equals(e.getName(), baseStateName))
                            .findFirst();
        assert parent.isPresent();
        if (!parent.isPresent())
            return;
        Optional<ESpecStateItem> child = parent.get().getChildren()
                            .stream()
                            .filter(e -> Objects.equals(e.getName(), subStateName))
                            .findFirst();
        assert child.isPresent();
        if (!child.isPresent())
            return;
        child.get().setStatus(newStatus);
    }
    public void addESpecLogItem(@Nonnull ESpecLogItem eSpecLogItem) {
        Platform.runLater(() -> eSpecLogItems.add(eSpecLogItem));
    }
    public void addESpecLogItem(@Nonnull String text, @Nonnull TaskStatusIndicator.Status progress){
        addESpecLogItem(new ESpecLogItem(text, progress));
    }
    public void addESpecLogItem(@Nonnull String text, boolean success){
        TaskStatusIndicator.Status progress;
        if (success) {
            progress = TaskStatusIndicator.Status.SUCCESS;
        } else {
            progress = TaskStatusIndicator.Status.FAILED;
        }
        addESpecLogItem(text, progress);
    }
    public void addESpecLogItem(@Nonnull String text, @Nonnull ESpecStepStatus status){
        addESpecLogItem(text, status.isSuccess());
    }




    @Override
    public void onPreFileLoad() {

    }

    @Override
    public void onPostFileLoad(@Nonnull FileSource fileSource,
                               boolean isFast, long byteSize,
                               @Nonnull ESpecStepStatus status) {
        if (status.isFailure())
            addESpecLogItem("File load \""+fileSource+"\"", status);
    }

    @Override
    public void onPostFileLoadAll(boolean success) {
        addESpecLogItem("All files loaded", success);
    }

    @Override
    public void onPreRSpec() {

    }

    @Override
    public void onRequestRSpecKnown(@Nonnull RspecSpec rspecSpec, @Nonnull String requestRspec) {
        addESpecLogItem("Request RSpec Available", true);
    }

    @Override
    public void onPostRSpec(@Nonnull RspecSpec rspec, @Nonnull ESpecStepStatus status) {
        if (status.isFailure())
            addESpecLogItem("RSpec Provision", status);
    }

    @Override
    public void onPostRspecAll(boolean success, @Nonnull List<String> allNodeClientIds) {
    }

    @Override
    public void onPreDir() {
        addSubESpecStatusItem(ExperimentState.SETUP_SOFTWARE.name(), new ESpecStateItem("Dir", ESpecStateItem.ESpecStateStatus.RUNNING));
    }

    @Override
    public void onPreDirNode(@Nonnull String nodeClientId) {

    }

    @Override
    public void onPostDir(@Nonnull DirSpec dir,
                          @Nonnull String fullPath,
                          @Nonnull String nodeClientId,
                          @Nonnull ESpecStepStatus status) {
        addESpecLogItem("Dir \""+fullPath+"\" Creation at "+nodeClientId, status);
    }

    @Override
    public void onPostDirNode(@Nonnull String nodeClientId, boolean success) {
        addESpecLogItem("Finished Dir Creations at "+nodeClientId, success);
    }

    @Override
    public void onPostDirAll(boolean success) {
        addESpecLogItem("Finished Dir Creations", success);
        setSubStatus(ExperimentState.SETUP_SOFTWARE.name(),
                "Dir",
                success ? ESpecStateItem.ESpecStateStatus.SUCCESS : ESpecStateItem.ESpecStateStatus.FAILURE);
    }

    @Override
    public void onPreUpload() {
        addSubESpecStatusItem(ExperimentState.SETUP_SOFTWARE.name(), new ESpecStateItem("Upload", ESpecStateItem.ESpecStateStatus.RUNNING));
    }

    @Override
    public void onPreUploadNode(@Nonnull String nodeClientId) {

    }

    @Override
    public void onPostUpload(@Nonnull UploadLikeSpec uploadSpec,
                             @Nonnull String fullPath,
                             @Nonnull String nodeClientId,
                             @Nonnull ESpecStepStatus status) {
        addESpecLogItem("Upload \""+fullPath+"\" at "+nodeClientId, status);
    }

    @Override
    public void onPostUploadNode(@Nonnull String nodeClientId, boolean success) {
        addESpecLogItem("Finished Uploads at "+nodeClientId, success);
    }

    @Override
    public void onPostUploadAll(boolean success) {
        addESpecLogItem("Finished Uploads", success);
        setSubStatus(ExperimentState.SETUP_SOFTWARE.name(),
                "Upload",
                success ? ESpecStateItem.ESpecStateStatus.SUCCESS : ESpecStateItem.ESpecStateStatus.FAILURE);
    }

    @Override
    public void onPreExecute() {
        addSubESpecStatusItem(ExperimentState.SETUP_SOFTWARE.name(), new ESpecStateItem("Execute", ESpecStateItem.ESpecStateStatus.RUNNING));
    }

    @Override
    public void onPostExecute(@Nonnull ExecuteSpec executeSpec,
                              @Nonnull String fullExecutablePath,
                              @Nullable String fullLogPath,
                              @Nonnull String nodeClientId,
                              int exitStatus,
                              @Nonnull ESpecStepStatus status) {
        addESpecLogItem("Execute \""+fullExecutablePath+"\" at "+nodeClientId, status);
    }

    @Override
    public void onPostExecuteStepAll(@Nonnull ExecuteSpec executeSpec, boolean success) {

    }

    @Override
    public void onPostExecuteAll(boolean success) {
        addESpecLogItem("Finished Executes", success);
        setSubStatus(ExperimentState.SETUP_SOFTWARE.name(),
                "Execute",
                success ? ESpecStateItem.ESpecStateStatus.SUCCESS : ESpecStateItem.ESpecStateStatus.FAILURE);
    }

    @Override
    public void onPreAnsible() {
        addSubESpecStatusItem(ExperimentState.SETUP_SOFTWARE.name(), new ESpecStateItem("Ansible", ESpecStateItem.ESpecStateStatus.RUNNING));
    }

    @Override
    public void onPostAnsiblePlaybook(@Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                      @Nonnull String fullPlaybookPath,
                                      @Nullable String fullLogPath,
                                      @Nonnull String nodeClientId,
                                      @Nonnull ESpecStepStatus status) {
        addESpecLogItem("Ansible Playbook \""+fullPlaybookPath+"\" at "+nodeClientId, status);
    }

    @Override
    public void onPostAnsibleAll(boolean success) {
        addESpecLogItem("Ansible Playbooks", success);
        setSubStatus(ExperimentState.SETUP_SOFTWARE.name(),
                "Ansible",
                success ? ESpecStateItem.ESpecStateStatus.SUCCESS : ESpecStateItem.ESpecStateStatus.FAILURE);
    }
}
